#!/usr/bin/env python

# --------------------------------------------------------------------------------
#                     type here YOUR NAME and your student ID
# -------------------------------------------------------------------------------

import rospy
import math
import numpy as np
from copy import deepcopy
from numpy.linalg import inv
from nav_msgs.msg import Odometry
from geometry_msgs.msg import PointStamped
from tf.transformations import euler_from_quaternion

def get_rotation (msg):
    global roll, pitch, yaw
    orientation_q = msg.pose.pose.orientation
    orientation_list = [orientation_q.x, orientation_q.y, orientation_q.z, orientation_q.w]
    (roll, pitch, yaw) = euler_from_quaternion (orientation_list)
    return yaw

def get_quaternion_from_euler(roll, pitch, yaw):

    cr = np.cos(roll * 0.5);
    sr = np.sin(roll * 0.5);
    cp = np.cos(pitch * 0.5);
    sp = np.sin(pitch * 0.5);
    cy = np.cos(yaw * 0.5);
    sy = np.sin(yaw * 0.5);

    qw = cr * cp * cy + sr * sp * sy;
    qx = sr * cp * cy - cr * sp * sy;
    qy = cr * sp * cy + sr * cp * sy;
    qz = cr * cp * sy - sr * sp * cy;

    return [qx, qy, qz, qw]

dt = 0.1 # Approximate time difference between measurements

sGPS     = 0.0001
sCourse  = 0.0001
sVelocity= 0.0001
sYaw     = 0.0001

# Process noise
Q = np.diag([sGPS**2, sGPS**2, sCourse**2, sVelocity**2, sYaw**2])

varGPS = 0.25   # Standard Deviation of GPS Measurement
varSpeed = 0.03 # Variance of the speed measurement
varYaw = 0.03   # Variance of the yawrate measurement

# Measurement Noise matrix
R = ???

# x: state vector, u: control vector
def motion_model(x, u):

    # the motion model describes the pose change of the robot given the current pose and control commands
    # the prediction step is based on the motion model of the robot taking into account the control inputs
    
    g = ???

    return g

# x: state vector
def jacob_g(x):

    # Jacobian matrix which is needed to propagate the covariances in the linearized motion model of the EKF
    Jg = ???

    return Jg


def ekf_estimation(xEst, Pest, z, u):

    # Prediction step
    # Predicted state estimate (mean)
    xPred = motion_model(xEst, u)

    # Jacobian matrix of g (motion model) with respect to state vector x
    Jg = jacob_g(xPred)
    
    # Predicted covariance estimate (Covariance matrix)
    Pest = ???
    
    # Correction step
    # h(X), create a vector from the estimated state with elments only the quantities that are considered that we measure
    hx = ???

    # Jacobian of mapping matrix h(X)
    # similar to matrix C of the KF 
    Jh = ???
  
    # Kalman Gain
    K = ???

    # Measurement matrix
    Z = np.empty((4,1))
    Z[0,0] = ???
    Z[1,0] = ???
    Z[2,0] = ???
    Z[3,0] = ???

    # EKF correction, updated state estimate (mean)
    xEst = ???

    # Updated covariance estimate
    Pest = ???

    return xEst,Pest


if __name__ == "__main__":
    # Initialize node
    rospy.init_node('EKF', anonymous=True)

    # ROS refresh rate
    rate = rospy.Rate(1/dt) # hz

    # Create a publisher that will publish your ekf estimation at /ekf topic using Odometry message
    ekf_pub = ???
    ekf_est = ??? # Define the message

    # To visualize the noisy gps measurements
    gps_pub = rospy.Publisher('/noisy_gps',PointStamped,queue_size=1)
    noisy_gps_msg = PointStamped()
    noisy_gps_msg.header.frame_id = "odom"

    # Create variables, use the numpy.empty function
    xEst  = ??? # state vector
    u     = ??? # control input
    z     = ??? # measurements
    Pest  = ??? # Covariance matrix
    

    initial_state_msg = rospy.wait_for_message("/odom",Odometry,timeout=None)
    dir0 = get_rotation(initial_state_msg)
    # Initialize the estimation to the current position
    xEst[0,0] = initial_state_msg.pose.pose.position.x + np.random.normal(0,varGPS)
    xEst[1,0] = initial_state_msg.pose.pose.position.y + np.random.normal(0,varGPS)
    xEst[2,0] = dir0
    xEst[3,0] = initial_state_msg.twist.twist.linear.x  + np.random.normal(0,varSpeed)
    xEst[4,0] = initial_state_msg.twist.twist.angular.z + np.random.normal(0,varYaw)

    # Kalman filter consists of a prediction and a correction step executed in a loop
    while not rospy.is_shutdown():

        input_msg = rospy.wait_for_message("/odom",Odometry,timeout=None)

        # Control input from odom topic
        # linear velocity
        u[0,0] = ???  + np.random.normal(0,varSpeed) # noise to linear velocity
        # angular velocity
        u[1,0] = ???  + np.random.normal(0,varYaw)   # noise to angular velocity

        # Measurement from odom topic
        # gps_x
        z[0,0] = ??? + np.random.normal(0,varGPS)    # Noisy measurement
        # gps_y
        z[1,0] = ??? + np.random.normal(0,varGPS)    # Noisy measurement

        xEst,Pest = ekf_estimation(xEst,Pest,z,u)

        # Visualize ekf estimation
        ekf_est = input_msg
        ekf_est.pose.pose.position.x = xEst[0,0]
        ekf_est.pose.pose.position.y = xEst[1,0]
        qx,qy,qz,qw = get_quaternion_from_euler(0,0,xEst[2,0])
        ekf_est.pose.pose.orientation.x = qx
        ekf_est.pose.pose.orientation.y = qy
        ekf_est.pose.pose.orientation.z = qz
        ekf_est.pose.pose.orientation.w = qw
        ekf_pub.publish(ekf_est)

        # Visualize noisy GPS
        noisy_gps_msg.point.x = z[0,0]
        noisy_gps_msg.point.y = z[1,0]
        noisy_gps_msg.point.z = 0
        gps_pub.publish(noisy_gps_msg)

        rate.sleep()
